%% Main function of NAG-FS framework for a fast and accurate classification.
% Details can be found in the original paper: https://www.scieC2edirect.com/scieC2e/article/abs/pii/S1361841519301367
% Islem Mhiri and Islem Rekik. "Joint input_ brain network atlas estimation and feature selection for neurological disorder diagnosis
%with application to autism",Medical Image Analysis, 2019, p. 101596.


%   ---------------------------------------------------------------------

%     This file contains the implementation of three key steps of our NAG-FS framework:
%     (1) Estimation of a centered and representative input_ network atlas,
%     (2) Discriminative connectional biomarker identification and  (3) disease classification:
%
%                        [AC1,AC2,ind] = NAG_FS(train_data,train_Labels,Nf,displays)
%
%                 Inputs:
%
%                          train_data: ((n-1) ? m ? m) tensor stacking the symmetric matrices of the training subjects
%                                      n the total number of subjects
%                                      m the number of nodes
%
%                          train_Labels: ((n-1) ? 1) vector of training labels (e.g., -1, 1)
%
%                          Nf: Number of selected features
%
%                          displays: Boolean variables [0, 1].
%                                    if displays = 1 ==> display(Atlas of group 1, Atlas of group 2, top features matrix and the circular graph)
%                                    if displays = 0 ==> no display
%                 Outputs:
%                         AC1: (m ? m) matrix storing the atlas of group 1
%                         洢һͼ׵ľ
%                        
%
%                         AC2: (m ? m) matrix storing the atlas of group 2
%
%                         ind: (Nf ? 1) vector storing the indices of the top disciminative features
%
%
%     To evaluate our framework we used Leave-One-Out cross validation strategy.



%To test NAG-FS on random data, we defined the function 'simulateData' where the size of the dataset is chosen by the user.
% ---------------------------------------------------------------------
%     Copyright 2019 Islem Mhiri, Sousse University.
%     Please cite the above paper if you use this code.
%     All rights reserved.
%     """

%%------------------------------------------------------------------------------


 
 


 function [AC1,AC2,ind,D2,Dif] = NAG_FS(train_data,train_Labels,Nf,displays)
  
 %% Step 1: Estimation of a centered and representative input_ network atlas
 %һĺʹԵͼ
 % Extract C1 group and C2 group 
 s1 = 0;
 s2 = 0;
 [~,~,sz3] = size(train_data);
 for h = 1: length(train_Labels)
     
     if (train_Labels(h) == 1)
         s1 = s1+1;
         XC1(s1,:,:) = squeeze(train_data(h,:,:));
     else
         s2 = s2+1;
         XC2(s2,:,:) = squeeze(train_data(h,:,:));
     end
     
 end
 
 
 %Disentangling the heterogeneous distribution of the input_ networks using SIMLR clustering method
 %simlr෨ֽ칹ֲ
 % C1 group
 k = [];
 
 for l = 1: s1
     k1 = squeeze((XC1(l,:,:)));
     k2 = k1(:); %vectorize the matrix ʸ
     k = [k;k2'];
 end
 
[t, S1, F1, ydata1,alpha1] = SIMLR(k,3,2);
 
 %C2 group
 kk = [];
 
 for l = 1: s2
     kk1 = squeeze(abs(XC2(l,:,:)));
     kk2 = kk1(:); %vectorize the matrix
     kk = [kk;kk2'];
 end
 [t1, S2, F2, ydata2,alpha2] = SIMLR(kk,3,2);
 
 % After using SIMLR, we extract each cluster independently for both classes
 %ʹSIMLR֮ǷֱΪȡÿȺ
 
 % C1 group
 qC11 = 1;
 qC12 = 1;
 qC13 = 1;
 
 for qC1 = 1: s1
     
     if t(qC1) == 1
         Ca1(qC11,:,:) = abs(XC1(qC1,:,:));
         qC11 = qC11+1;  %%%%%tм1
     elseif t(qC1) == 2
         Ca2(qC12,:,:) = abs(XC1(qC1,:,:));
         qC12 = qC12+1;   %%%%%%%tм2
     elseif t(qC1) == 3
         Ca3(qC13,:,:) = abs(XC1(qC1,:,:));
         qC13 = qC13+1;   %%%%%%tм3
     end
     
 end
 
 %C2 group
 qC21 = 1;
 qC22 = 1;
 qC23 = 1;
 
 for qC2 = 1: s2
     
     if t1(qC2) == 1
         Cn1(qC21,:,:) = (XC2(qC2,:,:));
         qC21 = qC21+1;   %%%%%t1м1
     elseif t1(qC2) == 2
         Cn2(qC22,:,:) = (XC2(qC2,:,:));
         qC22 = qC22+1;  %%%%%%%%t1м2
     elseif t1(qC2) == 3
         Cn3(qC23,:,:) = (XC2(qC2,:,:));
         qC23 = qC23+1;  %%%%%%t1м3
     end
     
 end
 
%  % For each cluster, we non-linearly diffuse and fuse all networks into a local centered network atlas using SNF
% %  %ÿأʹSNFɢں絽һΪĵͼ
 % Setting all the parameters.
 K = 20;%number of neighbors, usually (10~30)
 alpha = 0.5; %hyperparameter, usually (0.3~0.8)
 T = 20; %Number of Iterations, usually (10~20)
 
 % C1 group
  aa=[];
  aa1=[];
  aa2=[];
  aa3=[];
 for l = 1: (qC11-1)
     
     ll = num2str(l);
     Datap1.(['datap',ll,'']) = Standard_Normalization(squeeze(Ca1(l,:,:)));
  
%      Distp1.(['distp',ll,'']) = dist2( Datap1.(['datap',ll,'']), Datap1.(['datap',ll,'']));%֮ƽ
%      Wp1.(['Wp1',ll,'']) = affinityMatrix(Distp1.(['distp',ll,'']), K, alpha); %󣬳һټһ
    SS1 = svd(Datap1.(['datap',ll,'']));
  
    aa = cat(1,aa,SS1);
   
 end
%   Wall1 =struct2cell(Wp1);
   % aa = aa(find(aa));%%%%%%ҳDезԪ
    aa = nonzeros(aa);
%   aa=diag(aa);
%   aa=aa';
%   aa = [aa(1:10),aa(29:38),aa(57:66),aa(85:94)];
%   aa=aa(:);

 %S1(find(S1==0))=[];
 
%  if qC11>2
%      AC11 = SNF(Wall1,K,T) ;% First local network atlas for C1 group
%  else
%      AC11 = squeeze(Ca1(1,:,:));
%  end
 bb=[];
 for l = 1: (qC12-1)
     ll = num2str(l);
     Datap2.(['datap',ll,'']) = Standard_Normalization(squeeze(Ca2(l,:,:)));
     SS2 = svd(Datap2.(['datap',ll,'']));
     bb=cat(1,bb,SS2);
 end
  %bb = bb(find(bb));%%%%%%ҳDезԪ
  bb = nonzeros(bb);
 cc=[];
 for l = 1: (qC13-1)
     ll = num2str(l);
     Datap3.(['datap',ll,'']) = Standard_Normalization(squeeze(Ca3(l,:,:)));
     SS3 = svd(Datap3.(['datap',ll,''])); 
     cc=cat(1,cc,SS3);
    
 end
  %cc = cc(find(cc));%%%%%%ҳDезԪ
  cc = nonzeros(cc);
% %  cc=diag(cc);
% cc=cc(:);


 % S3(find(S3==0))=[];
%  if qC13>2
%      AC13 = SNF(Wall3,K,T); % Third local network atlas for C1 group
%  else
%      AC13 = squeeze(Ca3(1,:,:));
%  end
% % % %  % C2 group
dd=[];
 for l = 1: (qC21-1)
     ll = num2str(l);
     Datan1.(['datan',ll,'']) = Standard_Normalization(squeeze(Cn1(l,:,:)));
      SS4 =svd( Datan1.(['datan',ll,'']));
      dd=cat(1,dd,SS4);
 end
   % dd = dd(find(dd));%%%%%%ҳDезԪ
  dd = nonzeros(dd);
 ee=[];
 for l = 1: (qC22-1)
     ll = num2str(l);
     Datan2.(['datan',ll,'']) = Standard_Normalization(squeeze(Cn2(l,:,:)));
      SS5 =svd(Datan2.(['datan',ll,'']));
      ee=cat(1,ee,SS5);
 end
    ee = ee(find(ee));%%%%%%ҳDезԪ
   %ee = nonzeros(ee);

ff=[];
 for l = 1: (qC23-1)
     ll = num2str(l);
     Datan3.(['datan',ll,'']) = Standard_Normalization(squeeze(Cn3(l,:,:)));
      SS6 = svd(Datan3.(['datan',ll,'']));
      ff=cat(1,ff,SS6);
 end
    %ff = ff(find(ff));%%%%%%ҳDезԪ
  ff = nonzeros(ff);
 
 AC11=[aa;bb;cc];
 AC21=[dd;ee;ff];
 
%  AC11=AC11';
%  AC21=AC21';
%  AC11 = [AC11(1:10),AC11(29:38),AC11(57:66),AC11(85:94),AC11(113:122),AC11(141:150),AC11(169:178),AC11(197:206)];
%  AC21 = [AC21(1:10),AC21(29:38),AC21(57:66),AC21(85:94),AC21(113:122),AC21(141:150),AC21(169:178)];
%  AC11=diag(AC11);
%  AC21=diag(AC21);
% [rdata,ldata] = size(AC11);
%  n=round((-1+sqrt(1+8*ldata))/2) ;
% AA = zeros(n,n);
% 
% 
%   A = AC11;  
% for ii=1:n
%     for j=ii:n
%         index=sum(n:-1:n-ii+2)+j-ii+1;  
%         AA(ii,j)=A(index); 
%        
%     end
% end


%  [~,Bdata] = size(AC21);
%  nn=floor((-1+sqrt(1+8*Bdata))/2);
%  BB = [];
%  

%    
% for ii=1:nn
%     for j=ii:nn
%         index=sum(nn:-1:nn-ii+2)+j-ii+1;  
%         BB(ii,j)=B(index); 
%     end

 AC11=diag(AC11);
 AC21=diag(AC21); 
[mm,nn]=size(AC11);
AC1=AC11;
AC2=imresize(AC21,[mm,nn]);

% [mm,nn]=size(AA);
% [U7,AC1,V7]=svd(AA);
% [U8,AC2,V8]=svd(imresize(BB,[mm,nn]));
 
 D = abs(AC1-AC2);% Difference between matrices
 %D = abs(AC1-AC2);% Difference between matrices
 [sizeac,~]=size(AC1); %% Step 2:  Discriminative connectional biomarker identification % % % % %
 %D = triu(D,1); %Upper triangular part of matrix
 D1 = D(find(D));%%%%%%ҳDезԪأЩԪصֵصD
%  D1=D1(randperm(numel(D1),1953));
 D1=D1(1:512);
 D1 = D1.';
 D2 = D1;
%  D2=D1;
% [AAA,BBB]=size(D1);
% D1=D1(1:10:AAA);

%      D2=D1(1:3:end);
%   D2 = sort(D2,'descend');% Ranking features 

 D2=D2(randperm(numel(D2),100));
 Dif = D2(1:Nf); % Extract the top Nf ranked features
%  tempD2=cat(1,D2,tempD2);
%  tempDif=cat(1,Dif,tempDif);
%  save tempD2  tempDif
 ind = [];
%  Dif = gather(Dif);
%  D1 =  gather(D1);
 for i = 1: Nf
     [~,~,pos] = intersect(Dif(i),D1); % Select the indices of top Nf ranked features
     ind = [ind;pos];
 end
 qq = [];
 qq1 = [];
 topFeatures =zeros(sz3);
%  if sz1>sz3
%      sz=sz1;
%  else
%      sz=sz3;
%  end
  for h = 1 : Nf
     X = ind(h);
     [q,q1] = map_index_to_position_in_matrix(X,sz3);
     qq = [qq;q];
     qq1 = [qq1;q1];
  end
 for h = 1 : Nf
     topFeatures(qq(i),qq1(i)) = D2(i);
     topFeatures(qq1(i),qq(i)) = D2(i);
 end
 
 %% Display Atlas 1, Atlas 2 and the circular graph of the top discriminative features
 %%%%%%%ʾͼ1ͼ2top frature 
 
 
     
   if (displays)
     
     figure
     imagesc(AC1),title('Atlas of group 1 ','Color','b') % Display Atlas 1 of each LOO-CV iteration
     
     pause(2)
     figure
     imagesc(AC2) ,title('Atlas of group 2 ','Color','b') % Display Atlas 2 of each LOO-CV iteration.
     
     pause(2)
     figure
     imagesc(topFeatures) ,title('Top discriminative features ','Color','b') % Display Atlas 2 of each LOO-CV iteration.
     
     pause(2)
     
     close()
     figure
     J = circularGraph(topFeatures), title('Top feature ')
     
     pause(2)
     
   end
 end
 
 
 
 
 
